'******************************************************************************
'*  io.spin - handle input and output requests sent by pm80
'*
'*  (c) Juergen Buchmueller <pullmoll@t-online.de>
'*
'* $Id: io.spin,v 1.4 2010-06-05 06:43:50 pm Exp $
'******************************************************************************
CON

	io_cmd_out		=	$01		' I/O emulator command output
	io_cmd_in		=	$02		' I/O emulator command input
	io_cmd_break		=	$03		' I/O emulator command break
	io_cmd_kbd		=	$04		' I/O emulator command read TRS-80 keyboard

	baud			=	115200		' baud rate
	SCRATCH_RAM		=	$4000
	FSEL_TOP		=	2
	FSEL_H			=	14

VAR
	long	cog
	long	io_command
	long	params[2]
	long	rc
	word	topy
	word	boty
	word	ndir
	word	row
	word	y0
	byte	key
	byte	curx
	byte	cury
	byte	getdir
	byte	de[32]					' directory entry
	byte	name[12+1]
	byte	buff[64]

OBJ
	fat	: "fatfs"
#ifdef TV
	video	: "video"
#endif TV
#ifdef VGA
	video	: "vga"
#endif
	kbd	: "keyboard"

PUB start(pins) : okay | i, dent
	stop

	screen_ptr := @videoram
	chr(12)
	str(string("********* TRS-80 emulator on the Parallax Propeller *********",13,10))

	' ***************************************************************
	' Start the FAT driver and the SD SPI object (must be first)
	'
	okay := \fat.start(byte[pins][2], byte[pins][3], byte[pins][4], byte[pins][5])
	if okay < 0
		failure(string("FAT + SD SPI"),okay)
	else
		success(string("FAT + SD SPI"))
	iobuff_ptr := fat.get_memory

#ifdef TV
	' ***************************************************************
	' Start the TRS80 TV video object
	'
	params[0] := @videoram				' 1K video RAM
	params[1] := byte[pins][9]			' video pin
	video.start(@params)
#endif TV

#ifdef VGA
	' ***************************************************************
	' Start the TRS80 VGA video object
	'
	video.start(byte[pins][9], @videoram, @sync)
#endif VGA

	' ***************************************************************
	' Start the TRS80 keyboard driver
	'
	okay := \kbd.startx(byte[pins][6], byte[pins][7], 0, 40)
	if okay < 0
		failure(string("Keyboard"),okay)
		return
	kbd_states := kbd.states_ptr

	' ***************************************************************
	' Start the I/O handler cog
	'
	dira_mask := 1 << byte[pins][8]			' set audio pin mask
	ctra_mode := %00110 << 26 | byte[pins][8]	' set to single ended DUTY mode
	okay := cog := COGNEW(@entry, @io_command) + 1
	if okay =< 0
		failure(string("I/O"),okay)
		return -1

	' ***************************************************************
	' mount the FAT, i.e. check the format of the SD media
	'
	okay := \fat.mount(0)
	if okay < 0
		failure(string("FAT mount"), okay)
		return -1

	rom_load

	' *********************************************************
	' file selection
	'
	waitcnt(CNT + 2*clkfreq)

	getdir := 1
	repeat y0 from FSEL_TOP - 1 to 15
		gotoxy(0, y0)
		chr(22)
	repeat while getdir
		gotoxy(0, FSEL_TOP - 1)
		str(string("Sorting directory...",22))
		readdir
		row := 0
		topy := 0
		gotoxy(0, FSEL_TOP - 1)
		str(string("Select a file to load ("))
		dec(boty)
		chr("/")
		dec(ndir)
		str(string("):",22))
		repeat
			waitcnt(cnt + clkfreq >> 4)
			repeat y0 from 0 to FSEL_H - 1
				rc := direntry_name(topy + y0, @buff)
				if rc < 0
					buff[0] := " "
					buff[1] := " "
					buff[2] := 0
				gotoxy(0, FSEL_TOP+y0)
				if y0 == row
					str(string("=> "))
				else
					str(string("   "))
				str(@buff)
				if y0 == row
					str(string(" <="))
				chr(22)

			if kbd.keystate($34)	' Down arrow
				if (row < FSEL_H - 1) and (row < boty)
					row++
				elseif topy < boty - (FSEL_H - 1)
					topy++

			if kbd.keystate($33)	' Up arrow
				if row > 0
					row--
				elseif topy > 0
					topy--

			if kbd.keystate($3b)	' Page down
				if (topy + FSEL_H + row) < boty
					topy += FSEL_H
				elseif boty => FSEL_H
					topy := boty - (FSEL_H - 1)
					row := boty - topy
				else
					topy := 0
					row := boty

			if kbd.keystate($3a)	' Page up
				if (topy - FSEL_H) > 0
					topy -= FSEL_H
				else
					topy := 0
					row := 0

			if kbd.keystate($30)	' Return
				repeat y0 from 0 to FSEL_H - 1
					gotoxy(0, FSEL_TOP+y0)
					chr(22)
				gotoxy(0, FSEL_TOP)
				getdir := 0
				rc := direntry_name(topy + row, @buff)
				if byte[@buff][0] == "<"
					chdir_entry(topy + row)
					getdir := 1
				quit
		
	' *********************************************************
	' load the selected file
	'
	cmd_load(@buff)

	str(string("Press <RETURN> to start "))
	str(@buff)
	crlf
	repeat until !kbd.keystate($30)
	repeat until kbd.keystate($30)

	return 0

PUB stop
'' Stop CPU simulator - frees a cog
	if cog
		cogstop(cog~ - 1)

PUB command_addr
'' Return the address of the command long
	return @io_command

PUB get_videoram
'' Return the address of the video RAM
	return @videoram

PRI gotoxy(x,y)
	curx := (x #> 0) <# 63
	cury := (y #> 0) <# 15

PRI chr(ch)
	case ch
		22:	' clear to end of line
			bytefill(screen_ptr + cury << 6 + curx, $20, 64 - curx)
		13:	' carriage return
			curx := 0
		12:	' form feed - clear screen
			longfill(screen_ptr, $20202020, 64*16/4)
			curx := 0
			cury := 0
		10:	' line feed
			cury += 1
			if cury > 15
				longmove(screen_ptr, screen_ptr + 64, 64*15/4)
				longfill(screen_ptr + 64*15, $20202020, 64/4)
				cury := 15
		9:	' tab
			repeat until (curx & 7) == 0
				chr(32)
		8:	' back space
			if curx > 0
				--curx
		other:
			byte[screen_ptr][cury << 6 + curx] := ch
			curx++
			if curx == 64
				chr(13)
				chr(10)

PRI str(src) | i
	repeat i from 0 to strsize(src) - 1
		chr(byte[src][i])

PRI dec(value) | _i
'' Print a decimal number
	result := 0
	if value < 0
		chr("-")
		if (value == NEGX)
			chr("2")
			value += 2_000_000_000
		-value
	_i := 1_000_000_000
	repeat 10
		if value => _i
			chr(value / _i + "0")
			value //= _i
			result~~
		elseif result or _i == 1
			chr("0")
		_i /= 10

PRI crlf
	chr(13)
	chr(10)

PRI dec_size(value)
	if value > 1048576
		dec(value/1048576)
		chr(".")
		dec(10*(value // 1048576) / 1048576)
		str(string("MB"))
	elseif value > 1024
		dec(value/1024)
		chr(".")
		dec(10*(value // 1024) / 1024)
		str(string("KB"))
	else
		dec(value)

PRI hex(value, digits)
'' Print a hexadecimal number

	value <<= (8 - digits) << 2
	repeat digits
		chr(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))

PRI bin(value, digits)
'' Print a binary number

	value <<= 32 - digits
	repeat digits
		chr((value <-= 1) & 1 + "0")

PRI chr_ptr(ch, dst)
	byte[dst++] := ch
	byte[dst] := 0
	return dst

PRI str_ptr(src, dst) | i
	repeat i from 0 to strsize(src) - 1
		byte[dst++] := byte[src][i]
	byte[dst] := 0
	return dst

PRI dec_ptr(value, dst) | _i
'' Print a decimal number
	result := 0
	if value < 0
		dst := chr_ptr("-", dst)
		if (value == NEGX)
			dst := chr_ptr("2", dst)
			value += 2_000_000_000
		-value
	_i := 1_000_000_000
	repeat 10
		if value => _i
			dst := chr_ptr(value / _i + "0", dst)
			value //= _i
			result~~
		elseif result or _i == 1
			dst := chr_ptr("0", dst)
		_i /= 10
	return dst

PRI dec_size_ptr(value, dst)
	if value > 1048576
		dst := dec_ptr(value/1048576, dst)
		dst := chr_ptr(".", dst)
		dst := dec_ptr(10*(value // 1048576) / 1048576, dst)
		dst := str_ptr(string("MB"), dst)
	elseif value > 1024
		dst := dec_ptr(value/1024, dst)
		dst := chr_ptr(".", dst)
		dst := dec_ptr(10*(value // 1024) / 1024, dst)
		dst := str_ptr(string("KB"), dst)
	else
		dst := dec_ptr(value, dst)
	return dst

PRI failure(who, val)
	str(who)
	str(string(" failed: "))
	dec(val)
	crlf

PRI success(who)
	str(who)
	str(string(" initialized, "))
	dec(free_cogs)
	str(string(" cogs free.", 13, 10))

PUB free_cogs : free | i, cogs[8], jmp_0
	jmp_0 := %010111_0001_1111_000000000_000000000
	repeat while (cogs[free] := cognew(@jmp_0, 0)) => 0 
		free++
	if free
		repeat i from 0 to free - 1
			cogstop(cogs[i])
	return free

PRI strncmp(dst, src, n) | i
	repeat i from 0 to n - 1
		if byte[dst][i] <> byte[src][i]
			return byte[dst][i] - byte[src][i]
	return 0

PRI readdir | rec, i, swapped
	rec := fat.opendir(@de)
	boty := 0
	ndir := 0
	repeat while rec => 0
		ndir++
		if byte[@de][fat#dent_name] <> $e5	' deleted entry?
			if (fat.match(@de, string("*.CMD")) or fat.isdir(@de)) and not fat.isdot(@de)
				fat.write_xmm_ram(@de, SCRATCH_RAM + boty * 32, 32)
				boty++
		rec := fat.readdir(rec, @de)
	' do a bubble sort of the filenames in memory
	swapped := 1
	repeat until swapped == 0
		swapped := 0
		repeat i from 0 to boty - 2
			' read two directory entries
			fat.read_xmm_ram(@buff, SCRATCH_RAM + i * fat#dent_bytes, 2*fat#dent_bytes)
			' and compare the names
			if strncmp(@buff, @buff + fat#dent_bytes, 8) > 0
				fat.write_xmm_ram(@buff, SCRATCH_RAM + (i + 1) * fat#dent_bytes, fat#dent_bytes)
				fat.write_xmm_ram(@buff + fat#dent_bytes, SCRATCH_RAM + i * fat#dent_bytes, fat#dent_bytes)
				swapped++

PRI direntry_name(n, buf) | i, temp[fat#dent_bytes/4]
	if n < boty
		fat.read_xmm_ram(@temp, SCRATCH_RAM + n * 32, 32)
		if (byte[@temp][fat#dent_attr] & fat#dent_attr_D) <> 0
			' directory
			byte[buf] := "<"
			fat.get_dent_name(@temp, buf + 1)
			i := strsize(buf)
			byte[buf + i + 0] := ">"
			byte[buf + i + 1] := 0
		else
			fat.get_dent_name(@temp, buf)
		return 0
	return -1

PRI chdir_entry(n) | temp[fat#dent_bytes/4]
	if n < boty
		fat.read_xmm_ram(@temp, SCRATCH_RAM + n * fat#dent_bytes, fat#dent_bytes)
		fat.chdir_dent(@temp)
		return 0
	return -1

PRI rom_load | fp, dent, i
' find the TRS80.ROM file and load it into the ROM region
'
	fp := \fat.fopen(string("TRS80.ROM"), "r")
	if fp < 0
		return -1
	repeat i from $0000 to $2e00 step $200
		if fat.fread(fp, iobuff_ptr, $200) == $200
			fat.write_xmm_ram(iobuff_ptr, i, $200)
	fat.fclose(fp)
	str(string("Loaded to $0000..$2FFF"))
	return 0

PRI open_file(fname_ptr) : dent | size, sec
	dent := fat.find_file(fname_ptr)
	if dent < 0
		str(fname_ptr)
		str(string(" not found.",13,10))
		return -1
	bytemove(@de, dent, 32)
	fat.get_dent_name(@de, @name)
	size := fat.get_dent_size(@de)
	sec := fat.get_dent_first_sector(@de)
	str(@name)
	str(string(9,"sector "))
	dec(sec)
	str(string(9,"size "))
	dec_size(size)
	crlf

PRI cmd_load(fname_ptr) | dent, sec, nsec, pos, state, badr, blen, bmin, bmax
	dent := open_file(fname_ptr)
	if dent < 0
		return -1
	fat.get_dent_name(dent, @name)
	sec := fat.get_dent_first_sector(@de)

	pos := $200
	state := 0
	bmin := $ffff
	bmax := $0000
	repeat
		if pos == $200
			if sec < 0
				quit
			nsec := fat.get_dent_next_sector(@de, sec)
			fat.read_sector(sec)
			sec := nsec
			pos := 0
		case state
			0:	' expecting a header
				case byte[iobuff_ptr][pos++]
					$01:	' CMD load block header
						state := 1
					$02:	' CMD entry point header
						state := 7
					other:	' CMD comment block
						state := 5

			1:	' expecting a block length
				blen := byte[iobuff_ptr][pos++]
				state := 2

			2:	' expecting address LSB
				blen--
				badr := byte[iobuff_ptr][pos++]
				state := 3

			3:	' expecting address MSB
				blen--
				badr += 256 * byte[iobuff_ptr][pos++]
				state := 4
				if blen =< 0
					blen += 256
				bmin := bmin <# badr

			4:	' expecting blen bytes of data
				if pos + blen =< $200
					fat.write_xmm_ram(iobuff_ptr + pos, badr, blen)
					badr += blen
					pos += blen
					blen := 0
				else
					fat.write_xmm_ram(iobuff_ptr + pos, badr, $200 - pos)
					badr += $200 - pos
					blen -= $200 - pos
					pos := $200
				if blen == 0
					bmax := bmax #> badr
					state := 0

			5:	' expecting a block length
				blen := byte[iobuff_ptr][pos++]
				if blen == 0
					blen := 256
				state := 6

			6:	' ignoring comment data
				if pos + blen <= $200
					pos += blen
					blen := 0
				else
					blen -= $200 - pos
					pos := $200
				if blen == 0
					state := 0

			7:	' expecting a block length
				blen := byte[iobuff_ptr][pos++]
				state := 8

			8:	' expecting entry point LSB
				blen--
				badr := byte[iobuff_ptr][pos++]
				state := 9

			9:	' expecting entry point MSB
				blen--
				badr += 256 * byte[iobuff_ptr][pos++]
				str(string("Loaded from "))
				hex(bmin, 4)
				str(string(".."))
				hex(bmax - 1, 4)
				crlf
				str(string("Entry point "))
				hex(badr, 4)
				crlf
				fat.write_xmm_ram(@badr, $0103, 2)
				quit

DAT
'---------------------------------------------------------------------------------------------------------

		org	0

'---------------------------------------------------------------------------------------------------------
entry
iocmd_ptr		mov	iocmd_ptr, PAR
			jmp	#run

screen_ptr		long	0-0
addr			long	0
cur_x			long	0
cur_y			long	0
iobuff_ptr		long	0-0
x20202020		long	$20202020
x10_000			long	10_000

dira_mask		long	0-0
ctra_mode		long	0-0
frqa_0			long	$23af_0000				' 0.46V of 1V, ie. 46%
frqa_1			long	$0000_0001				' 0V
frqa_2			long	$41f0_0000				' 0.85V of 1V, ie. 85%
frqa_3			long	$0000_0001				' 0V

kbd_states		long	0-0
kbd_index		long	0
kbd_shift		long	0
kbd_data		long	0

minus1			long	$ffffffff
low_24bits		long	$00ffffff
low_16bits		long	$0000ffff
msg_break_ptr		long	@@@msg_break

iocmd			long	0
port			long	0
data			long	0
null			long	0

t1			long	0
t2			long	0
t3			long	0
t4			long	0

strptr			long	0
hexval			long	0
hexcnt			long	0

port_ff			long	0

run
			wrlong	null, iocmd_ptr

			mov	DIRA, dira_mask			' enable the sound output
			mov	CTRA, ctra_mode			' set the counter mode to single ended DUTY mode
			mov	FRQA, frqa_0			' set the default output pulse width

command
			rdlong	iocmd, iocmd_ptr
			mov	t1, iocmd
			shr	t1, #24			WZ
		if_z	jmp	#command
			cmp	t1, #io_cmd_out		WZ
		if_z	jmp	#output
			cmp	t1, #io_cmd_in		WZ
		if_z	jmp	#input
			cmp	t1, #io_cmd_break	WZ
		if_z	jmp	#break
			cmp	t1, #io_cmd_kbd		WZ
		if_z	jmp	#key_matrix
command_0
			and	iocmd, low_24bits
			wrlong	iocmd, iocmd_ptr
			jmp	#command

'*******************************************
'
' Read i/o port and data to write to
'
output
			mov	port, iocmd			' get command word
			shr	port, #8			' extract port
			and	port, #$ff			' lower 8 bits
			mov	data, iocmd			' get command word
			and	data, #$ff			' extract data

			cmp	port, #$ff		WZ	' port = $FF?
		if_z	jmp	#output_ff

			jmp	#command_0


output_ff
			mov	port_ff, data			' save value
			mov	t1, data
			and	t1, #3
			add	t1, #frqa_0
			movs	:padr, t1
			mov	t1, data
:padr			mov	t2, 0-0				' get audio level
			mov	FRQA, t2			' write to FRQA
			' TODO: check the other bits of data
			jmp	#command_0

'*******************************************
'
' Read i/o port to read from
'
input
			andn	iocmd, #$ff			' clear the input data
			mov	port, iocmd			' get command word
			shr	port, #8
			and	port, #$ff

			cmp	port, #$ff		WZ	' port = $FF?
		if_z	jmp	#input_ff

			or	iocmd, #$ff			' return $ff for other ports
			jmp	#command_0

input_ff
			mov	data, port_ff
			andn	data, #%00010000		' clear input bit
			' TODO: sample a pin to read from cassette ?
			shl	data, #3
			and	data, #%11000000
			or	iocmd, data
			jmp	#command_0
		
'*******************************************
'
' Output a character to the screen
'
conout_chr
			mov	addr, cur_y			' y
			shl	addr, #6			' * 64
			add	addr, cur_x			' + x
			add	addr, screen_ptr		' + screen address

			cmp	data, #8		WZ	' backspace?
		if_z	jmp	#:bsp
			cmp	data, #9		WZ	' tab?
		if_z	jmp	#:tab
			cmp	data, #10		WZ	' line feed?
		if_z	jmp	#:lf
			cmp	data, #12		WZ	' clear screen?
		if_z	jmp	#:cls
			cmp	data, #13		WZ	' carriage return?
		if_z	jmp	#:cr
:out
			wrbyte	data, addr
			add	cur_x, #1
:wrap
			cmp	cur_x, #64		WZ
		if_nz	jmp	#:exit
			mov	cur_x, #0
:lf
			add	cur_y, #1
			cmp	cur_y, #16		WZ
		if_nz	jmp	#:exit
			call	#scroll
			mov	cur_y, #15
			jmp	#:exit

:bsp
			cmp	cur_x, #0		WZ, WC
		if_be	jmp	#:exit
			sub	cur_x, #1
			jmp	#:exit
:tab
			mov	data, #32
			wrbyte	data, addr
			add	addr, #1
			add	cur_x, #1
			test	cur_x, #7		WZ
		if_nz	jmp	#:tab
			jmp	#:wrap
:cr
			mov	cur_x, #0
			jmp	#:exit
:cls
			mov	addr, screen_ptr
			mov	cur_y, #64*16/4
:loop
			wrlong	x20202020, addr
			add	addr, #4
			djnz	cur_y, #:loop
			mov	cur_x, #0
:exit
conout_chr_ret
			ret		

'*******************************************
'
' Scroll the screen one line up
'
scroll
			mov	addr, screen_ptr
			mov	cur_y, #15
:row
			mov	cur_x, #64/4
:col
			add	addr, #64
			rdlong	data, addr
			sub	addr, #64
			wrlong	data, addr
			add	addr, #4
			djnz	cur_x, #:col
			djnz	cur_y, #:row
			mov	cur_x, #64/4
:fill
			wrlong	x20202020, addr
			add	addr, #4
			djnz	cur_x, #:fill
scroll_ret
			ret


'*******************************************
'
' Output a byte in hexval to the console
'
conout_hex_b
			mov	hexcnt, #2
			shl	hexval, #24
			jmp	#conout_hex

'*******************************************
'
' Output a word in hexval to the console
'
conout_hex_w
			mov	hexcnt, #4
			shl	hexval, #16
			jmp	#conout_hex

'*******************************************
'
' Output a long in hexval to the console
'
conout_hex_l
			mov	hexcnt, #8
conout_hex
			mov	data, hexval
			shr	data, #28
			cmp	data, #10	WZ, WC
		if_b	add	data, #48
		if_ae	add	data, #55
			call	#conout_chr
			shl	hexval, #4
			djnz	hexcnt, #conout_hex
conout_hex_ret
conout_hex_l_ret
conout_hex_w_ret
conout_hex_b_ret
			ret

'*******************************************
'
' Output a carriage return to the console
'
conout_space
			mov	data, #32
			call	#conout_chr
conout_space_ret
			ret

'*******************************************
'
' Output a carriage return to the console
'
conout_crlf
			mov	data, #13
			call	#conout_chr
			mov	data, #10
			call	#conout_chr
conout_crlf_ret
			ret

'*******************************************
'
' Output a string at strptr to the console
'
conout_string
			rdbyte	data, strptr	WZ
			add	strptr, #1
		if_z	jmp	#conout_string_ret
			call	#conout_chr
			jmp	#conout_string
conout_string_ret
			ret

'*******************************************
'
' Break - issued when the Z80 encounters
' a HALT instruction.
'
break
			mov	t1, iocmd			' read the address of the register file
			shr	t1, #8				' bits 23..8 to 15..0
			and	t1, low_16bits			' mask only bits 15..0
			mov	strptr, msg_break_ptr	

			mov	t2, #8
:loop_1			call	#conout_string
			rdword	hexval, t1			' read BC, DE, HL, AF, BC', DE', HL', AF'
			call	#conout_hex_w
			add	t1, #2
			djnz	t2, #:loop_1

			call	#conout_string			' new line and print "R:"
			rdbyte	hexval, t1			' R
			add	t1, #1
			rdbyte	t2, t1				' R2
			and	hexval, #%01111111		' R register's bit #7
			or	hexval, t2			' does never change from the last written value
			call	#conout_hex_b
			add	t1, #1

			call	#conout_string
			rdbyte	hexval, t1			' IFF1+2/IM
			mov	t2, t1				' save value
			and	hexval, #1			' IFF1 (interrupt flip-flop 1)
			call	#conout_hex_b

			call	#conout_string
			mov	hexval, t2			' IFF1+2/IM
			shr	hexval, #1			' shift bit #1 to #0
			and	hexval, #1			' IFF2 (interrupt flip-flop 2)
			call	#conout_hex_b

			call	#conout_string
			mov	hexval, t2			' IFF1+2/IM
			shr	hexval, #2			' shift bits #3..2 to #1..0
			and	hexval, #3			' IM (interrupt mode)
			call	#conout_hex_b
			add	t1, #1

			call	#conout_string
			rdbyte	hexval, t1			' I (interrupt vector register)
			call	#conout_hex_b
			add	t1, #1

			mov	t2, #4
:loop2			call	#conout_string
			rdword	hexval, t1			' IX, IY, SP, PC
			mov	t3, hexval
			call	#conout_hex_w
			add	t1, #2
			djnz	t2, #:loop2
			call	#conout_crlf

			jmp	#command_0

'*******************************************
'
' Read the TRS-80 keyboard matrix
' Address is in io_port, address bits 0..7
' determine the keyboard row(s) to read
'
key_matrix
			mov	port, iocmd			' get command word
			shr	port, #8
			and	port, low_16bits
			mov	t1, #0				' resulting bits
			mov	t2, kbd_states			' address of the keyboard state matrix
			mov	t3, #8				' 8 of them
			mov	data, #$01			' bit mask for the address lines
:loop			rdbyte	kbd_data, t2			' next keyboard row entry
			add	t2, #1
			test	port, data		WZ	' test port address A0, A1, A2 ...
		if_nz	or	t1, kbd_data			' if address bit was set, OR the data
			rol	data, #1			' rotate mask left
			djnz	t3, #:loop			' 8 times
			or	iocmd, t1			' return it in the command word bits 7..0
			jmp	#command_0

			fit	$1f0

sync			long	0
videoram		byte	0[64*16]

msg_break		byte	"BC:",0
			byte	" DE:",0
			byte	" HL:",0
			byte	" AF:",0
			byte	" BC'",0
			byte	" DE'",0
			byte	" HL'",0
			byte	" AF'",0
			byte	13,10, "R:",0
			byte	" IFF1:",0
			byte	" IFF2:",0
			byte	" IM:",0
			byte	" I:",0
			byte	" IX:",0
			byte	" IY:",0
			byte	" SP:",0
			byte	" PC:",0
			byte	" = ",0

''      MIT LICENSE
{{
'  Permission is hereby granted, free of charge, to any person obtaining
'  a copy of this software and associated documentation files
'  (the "Software"), to deal in the Software without restriction,
'  including without limitation the rights to use, copy, modify, merge,
'  publish, distribute, sublicense, and/or sell copies of the Software,
'  and to permit persons to whom the Software is furnished to do so,
'  subject to the following conditions:
'
'  The above copyright notice and this permission notice shall be included
'  in all copies or substantial portions of the Software.
'
'  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
'  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
'  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
'  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
'  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
'  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
'  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
}}
